# -*- coding: utf-8 -*-
"""Django plugin for guardrail.
Note: Define permission schema factory in `models.py` so that Django migrations
can detect permission schemas.
"""

from __future__ import absolute_import

from django import db

from guardrail.core import models
from guardrail.core import exceptions


class DjangoPermissionManager(models.BasePermissionManager):

    @staticmethod
    def _is_saved(record):
        """Django cannot create references to unsaved records."""
        return record.pk is not None

    @staticmethod
    def _build_query(query, agent, target, schema,
                     Agent=None, Target=None, custom=None):
        if Agent is None:
            query = query.filter(agent=agent)
        if Target is None:
            query = query.filter(target=target)
        if custom is not None:
            query = custom(query, agent=agent, target=target, schema=schema)
        return query

    def _get_permissions(self, agent, target, schema,
                         Agent=None, Target=None, custom=None):
        query = schema.objects.only('permission')
        query = self._build_query(
            query, agent, target, schema,
            Agent=Agent, Target=Target, custom=custom,
        )
        return {each.permission for each in query}

    def _has_permission(self, agent, target, schema, permission,
                        Agent=None, Target=None, custom=None):
        query = schema.objects.only('permission')
        query = query.filter(permission=permission)
        query = self._build_query(
            query, agent, target, schema,
            Agent=Agent, Target=Target, custom=custom,
        )
        return query.exists()

    def _add_permission(self, agent, target, schema, permission):
        try:
            return schema.objects.create(
                agent=agent,
                target=target,
                permission=permission,
            )
        except db.IntegrityError:
            raise exceptions.PermissionExists()

    def _remove_permission(self, agent, target, schema, permission):
        query = schema.objects.only('permission')
        query = query.filter(permission=permission)
        query = self._build_query(query, agent, target, schema)
        # Note: This emits an unnecessary extra query. This can be fixed once
        # patch once ticket 16891 is resolved.
        if not query.count():
            raise exceptions.PermissionNotFound
        query.delete()


class DjangoPermissionSchemaFactory(models.BasePermissionSchemaFactory):

    @staticmethod
    def _get_table_name(schema):
        return schema._meta.db_table

    def _make_schema_meta(self, agent, target):
        return type(
            'Meta',
            (object, ),
            dict(
                db_table=self._make_table_name(agent, target),
                unique_together=(('agent', 'target', 'permission'), )
            ),
        )

    def _make_schema_dict(self, agent, target):
        return dict(
            Meta=self._make_schema_meta(agent, target),
            __module__=__name__,
            id=db.models.AutoField(primary_key=True),
            agent=db.models.ForeignKey(agent, null=False, db_index=True),
            target=db.models.ForeignKey(target, null=False, db_index=True),
            permission=db.models.CharField(max_length=255, null=False, db_index=True),
        )


class DjangoLoader(models.BaseLoader):

    def __init__(self, schema, column='pk', kwarg='id'):
        super(DjangoLoader, self).__init__(schema, column, kwarg)

    def __call__(self, *args, **kwargs):
        query = {self.column: kwargs.get(self.kwarg)}
        return self.schema.objects.filter(**query).first()



# -*- coding: utf-8 -*-

from django.http import HttpResponse
from django.contrib.auth.decorators import login_required


@login_required
def index(request):
    return HttpResponse('welcome to zombocom')



# -*- coding: utf-8 -*-
"""Pony plugin for guardrail."""

from __future__ import absolute_import

import pony.orm as pn

from guardrail.core import models
from guardrail.core import exceptions


class PonyPermissionManager(models.BasePermissionManager):

    @staticmethod
    def _is_saved(record):
        return True

    @staticmethod
    def _build_query(query, agent, target, schema,
                     Agent=None, Target=None, custom=None):
        if Agent is None:
            query = query.filter(lambda row: row.agent == agent)
        if Target is None:
            query = query.filter(lambda row: row.target == target)
        if custom is not None:
            query = custom(query, agent=agent, target=target, schema=schema)
        return query

    def _get_permissions(self, agent, target, schema,
                         Agent=None, Target=None, custom=None):
        query = pn.select(row for row in schema)
        query = self._build_query(
            query, agent, target, schema,
            Agent=Agent, Target=Target, custom=custom,
        )
        return {row.permission for row in query}

    def _has_permission(self, agent, target, schema, permission,
                        Agent=None, Target=None, custom=None):
        query = pn.select(row for row in schema if row.permission == permission)
        query = self._build_query(
            query, agent, target, schema,
            Agent=Agent, Target=Target, custom=custom,
        )
        return query.exists()

    def _add_permission(self, agent, target, schema, permission):
        try:
            return schema(
                agent=agent,
                target=target,
                permission=permission,
            )
        except pn.CacheIndexError:
            raise exceptions.PermissionExists()

    def _remove_permission(self, agent, target, schema, permission):
        query = pn.select(row for row in schema if row.permission == permission)
        query = self._build_query(query, agent, target, schema)
        row = query.first()
        if not row:
            raise exceptions.PermissionNotFound
        row.delete()


class PonyPermissionSchemaFactory(models.BasePermissionSchemaFactory):

    @staticmethod
    def _get_table_name(schema):
        return schema._table_ or schema.__name__

    def _update_schema(self, schema, name, reverse):
        type.__setattr__(schema, name, reverse)
        reverse._init_(schema, name)
        schema._attrs_.append(reverse)
        schema._new_attrs_.append(reverse)

    def _update_agent(self, agent, target, schema):
        reverse = pn.Set(schema)
        name = 'targets_{0}'.format(self._get_table_name(target).lower())
        self._update_schema(agent, name, reverse)

    def _update_target(self, agent, target, schema):
        reverse = pn.Set(schema)
        name = 'agents_{0}'.format(self._get_table_name(agent).lower())
        self._update_schema(target, name, reverse)

    def _update_parents(self, agent, target, schema):
        self._update_agent(agent, target, schema)
        self._update_target(agent, target, schema)

    def _make_schema_dict(self, agent, target):
        return dict(
            _table_=self._make_table_name(agent, target),
            _indexes_=[
                pn.core.Index(
                    'agent', 'target', 'permission',
                    is_pk=False, is_unique=True,
                )
            ],
            id=pn.PrimaryKey(int, auto=True),
            agent=pn.Required(agent, index=True),
            target=pn.Required(target, index=True),
            permission=pn.Required(str, 255, index=True),
        )


class PonyLoader(models.BaseLoader):

    def __init__(self, schema, column=None, kwarg='id'):
        column = column if column is not None else schema._pk_.name
        super(PonyLoader, self).__init__(schema, column, kwarg)

    def __call__(self, *args, **kwargs):
        query = {self.column: kwargs.get(self.kwarg)}
        return pn.select(
            row for row in self.schema
        ).filter(
            **query
        ).first()



# -*- coding: utf-8 -*-

import flask
from flask.ext.login import LoginManager

import models
import fixtures
import permissions

app = flask.Flask(__name__)
app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///test.db'
app.config['SECRET_KEY'] = 'secret'

login_manager = LoginManager(app)
models.db.init_app(app)
with app.app_context():
    models.db.create_all()

fixtures.ensure(app)

@login_manager.user_loader
@login_manager.request_loader
def load_user_from_request(*args, **kwargs):
    return models.db.session.query(models.User).get(1)

@app.route('/posts/<id>/', methods=['GET'])
@permissions.has_post_permission('read')
def view_post(agent, target, **kwargs):
    return flask.jsonify(
        title=target.title,
        content=target.content,
    )

@app.route('/posts/<id>/', methods=['PUT'])
@permissions.has_post_permission('write')
def edit_post(agent, target, **kwargs):
    data = flask.request.get_json()
    target.content = data['content']
    models.db.session.add(target)
    models.db.session.commit()
    return flask.jsonify(status='success')



# -*- coding: utf-8 -*-

import functools


AGENT_NOT_FOUND = 'agent_not_found'
TARGET_NOT_FOUND = 'target_not_found'
FORBIDDEN = 'forbidden'


class has_permission(object):
    """Factory for permission-checking decorators. Should be subclassed or
    have arguments pre-filled with `functools.partial`.

    :param permission: Permission value
    :param manager: Permission manager
    :param agent_loader: Callable that loads an agent from the `args` and `kwargs`
        passed to the decorated function
    :param target_loader: Callable that loads a target from the `args` and `kwargs`
        passed to the decorated function
    :param error_handler: Callable that handles error codes `AGENT_NOT_FOUND`,
        `TARGET_NOT_FOUND`, and `FORBIDDEN`
    """
    def __init__(self, permission, manager, agent_loader, target_loader, error_handler):
        self.permission = permission
        self.manager = manager
        self.agent_loader = agent_loader
        self.target_loader = target_loader
        self.error_handler = error_handler

    def __call__(self, func):
        """Decorator that checks permissions before calling `func`. Note: wrapped
        function will be called with the loaded agent and target.

        :param func: Callable to decorate
        """
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            agent, target = self._check_permission(*args, **kwargs)
            kwargs.update({
                'agent': agent,
                'target': target,
            })
            return func(*args, **kwargs)
        return wrapped

    def _check_permission(self, *args, **kwargs):
        """Load agent and target records from `agent_loader` and `target_loader`,
        then check for the requested permission. Call `error_handler` if either
        loader returns `None`, or if permission is not present.
        """
        agent = self.agent_loader(*args, **kwargs)
        if not agent:
            return self.error_handler(AGENT_NOT_FOUND)
        target = self.target_loader(*args, **kwargs)
        if not target:
            return self.error_handler(TARGET_NOT_FOUND)
        if not self.manager.has_permission(agent, target, self.permission):
            self.error_handler(FORBIDDEN)
        return agent, target



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import sys
sys.path.append('/Users/jmcarp/code/guardian')

from django.db import models as dj
from django.contrib.auth.models import AbstractUser

from guardrail.core.registry import registry

from guardrail.ext.django.models import DjangoPermissionSchemaFactory


@registry.agent
class RoleGroup(dj.Model):
    pass


class Role(dj.Model):
    role_groups = dj.ManyToManyField(RoleGroup)


class User(AbstractUser):
    roles = dj.ManyToManyField(Role)


@registry.target
class Project(dj.Model):
    pass


@registry.target
class File(dj.Model):
    pass


class Label(dj.Model):
    file = dj.ForeignKey(File, related_name='labels')


factory = DjangoPermissionSchemaFactory((dj.Model, ))
registry.make_schemas(factory)



# -*- coding: utf-8 -*-
"""Custom object permissions backend for Django plugin"""

from .models import DjangoPermissionManager


class ObjectPermissionBackend(object):
    """Custom authentication backend for object-level permissions. Must be used
    in conjunction with a backend that handles the `authenticate` and `get_user`
    methods, such as the default `django.contrib.auth.backends.ModelBackend`.
    """
    def authenticate(self, username=None, password=None, *kwargs):
        return None

    def get_user(self, user_id):
        return None

    def has_perm(self, user, perm, target=None):
        manager = DjangoPermissionManager()
        return manager.has_permission(user, target, perm)

    def get_all_permissions(self, user, target=None):
        manager = DjangoPermissionManager()
        return manager.get_permissions(user, target)



# -*- coding: utf-8 -*-
"""SQLAlchemy plugin for guardrail."""

from __future__ import absolute_import

import sqlalchemy as sa

from guardrail.core import models
from guardrail.core import exceptions
from guardrail.core.registry import registry


def _get_primary_column(schema):
    primary = sa.inspection.inspect(schema).primary_key
    if len(primary) == 1:
        return primary[0]
    raise RuntimeError('Composite foreign keys not currently supported')


class SqlalchemyPermissionManager(models.BasePermissionManager):

    def __init__(self, session, registry=registry):
        super(SqlalchemyPermissionManager, self).__init__(registry)
        self.session = session

    @staticmethod
    def _is_saved(record):
        return True

    @staticmethod
    def _build_query(query, agent, target, schema, Agent=None, Target=None, custom=None):
        if Agent is None:
            query = query.filter(schema.agent == agent)
        if Target is None:
            query = query.filter(schema.target == target)
        if custom is not None:
            query = custom(query, agent=agent, target=target, schema=schema)
        return query

    def _get_permissions(self, agent, target, schema,
                         Agent=None, Target=None, custom=None):
        query = self.session.query(schema.permission)
        query = self._build_query(
            query, agent, target, schema,
            Agent=Agent, Target=Target, custom=custom,
        )
        return {each.permission for each in query}

    def _has_permission(self, agent, target, schema, permission,
                        Agent=None, Target=None, custom=None):
        query = self.session.query(schema.permission)
        query = query.filter(schema.permission == permission)
        query = self._build_query(
            query, agent, target, schema,
            Agent=Agent, Target=Target, custom=custom,
        )
        return bool(query.first())

    def _add_permission(self, agent, target, schema, permission):
        row = schema(
            agent=agent,
            target=target,
            permission=permission,
        )
        self.session.add(row)
        try:
            self.session.flush()
        except sa.exc.IntegrityError:
            self.session.rollback()
            raise exceptions.PermissionExists()
        return row

    def _remove_permission(self, agent, target, schema, permission):
        query = self.session.query(schema)
        query = query.filter(schema.permission == permission)
        query = self._build_query(query, agent, target, schema)
        count = query.delete()
        if not count:
            raise exceptions.PermissionNotFound


def _reference_column(schema, **kwargs):
    return sa.Column(
        sa.Integer,
        sa.ForeignKey(
            _get_primary_column(schema),
            onupdate='CASCADE',
            ondelete='CASCADE',
        ),
        **kwargs
    )


class SqlalchemyPermissionSchemaFactory(models.BasePermissionSchemaFactory):
    """Permission schema factory for use with SQLAlchemy.

    :param tuple bases: Base classes for created schema classes
    :param bool cascade: Database backend supports `ON DELETE` and `ON UPDATE`
        cascades; see :meth:`_update_parents` for details
    """
    def __init__(self, bases, cascade=False):
        super(SqlalchemyPermissionSchemaFactory, self).__init__(bases)
        self.cascade = cascade

    @staticmethod
    def _get_table_name(schema):
        return schema.__tablename__

    def _update_parents(self, agent, target, schema):
        """Create a many-to-many `relationship` between the `agent` and `target`
        schemas, using the created `schema` as the join table.

        Note: Use the `passive_deletes` and `passive_updates` flags only if the
        database backend supports the ON UPDATE and ON DELETE cascades. These
        options are not supported in SQLite, or in MySQL using the MyISAM storage
        engine.
        """
        attr = 'targets_{0}'.format(schema.__tablename__)
        backref = 'agents_{0}'.format(schema.__tablename__)
        relation = sa.orm.relationship(
            target,
            secondary=schema.__table__,
            backref=backref,
            passive_deletes=self.cascade,
            passive_updates=self.cascade,
        )
        setattr(agent, attr, relation)

    def _make_schema_dict(self, agent, target):
        return dict(
            __tablename__=self._make_table_name(agent, target),
            __table_args__=(
                sa.UniqueConstraint('agent_id', 'target_id', 'permission'),
            ),
            id=sa.Column(sa.Integer, primary_key=True),
            agent_id=_reference_column(agent, nullable=False, index=True),
            agent=sa.orm.relationship(agent),
            target_id=_reference_column(target, nullable=False, index=True),
            target=sa.orm.relationship(target),
            permission=sa.Column(sa.String, nullable=False, index=True),
        )


class SqlalchemyLoader(models.BaseLoader):

    def __init__(self, schema, session, column=None, kwarg='id'):
        column = column if column is not None else _get_primary_column(schema)
        super(SqlalchemyLoader, self).__init__(schema, column, kwarg)
        self.session = session

    def __call__(self, *args, **kwargs):
        return self.session.query(
            self.schema
        ).filter(
            self.column == kwargs.get(self.kwarg)
        ).first()



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import httplib
import functools

import flask
from flask.ext.login import current_user

from guardrail.core import decorators
from guardrail.ext.sqlalchemy import SqlalchemyLoader
from guardrail.ext.sqlalchemy import SqlalchemyPermissionManager

import models

manager = SqlalchemyPermissionManager(models.db.session)

def user_loader(*args, **kwargs):
    return current_user._get_current_object()

error_messages = {
    decorators.AGENT_NOT_FOUND: httplib.NOT_FOUND,
    decorators.TARGET_NOT_FOUND: httplib.NOT_FOUND,
    decorators.FORBIDDEN: httplib.FORBIDDEN,
}
def error_handler(message):
    flask.abort(error_messages.get(message, httplib.INTERNAL_SERVER_ERROR))

has_permission = functools.partial(
    decorators.has_permission,
    manager=manager,
    agent_loader=user_loader,
    error_handler=error_handler,
)

has_post_permission = functools.partial(
    has_permission,
    target_loader=SqlalchemyLoader(models.Post, models.db.session)
)



# -*- coding: utf-8 -*-

import models
import permissions


def ensure(app):
    with app.app_context():
        user = models.db.session.query(models.User).get(1)
        if not user:
            user = models.User(id=1, username='freddie')
            models.db.session.add(user)
            models.db.session.commit()
        post = models.db.session.query(models.Post).get(1)
        if not post:
            post = models.Post(id=1, title='death on two legs', content='dedicated to...')
            models.db.session.add(post)
            models.db.session.commit()
        try:
            permissions.manager.ensure_permission(user, post, 'read')
        except:
            pass
        models.db.session.commit()



# -*- coding: utf-8 -*-
"""Peewee plugin for guardrail."""

from __future__ import absolute_import

import peewee as pw

from guardrail.core import models
from guardrail.core import exceptions


class PeeweePermissionManager(models.BasePermissionManager):

    @staticmethod
    def _is_saved(record):
        """Peewee cannot create references to unsaved records."""
        return record.get_id() is not None

    @staticmethod
    def _build_query(query, agent, target, schema, Agent=None, Target=None, custom=None):
        if Agent is None:
            query = query.where(schema.agent == agent)
        if Target is None:
            query = query.where(schema.target == target)
        if custom is not None:
            query = custom(query, agent=agent, target=target, schema=schema)
        return query

    def _get_permissions(self, agent, target, schema,
                         Agent=None, Target=None, custom=None):
        query = schema.select(schema.permission)
        query = self._build_query(
            query, agent, target, schema,
            Agent=Agent, Target=Target, custom=custom,
        )
        return {each.permission for each in query}

    def _has_permission(self, agent, target, schema, permission,
                        Agent=None, Target=None, custom=None):
        query = schema.select(schema.permission)
        query = query.where(schema.permission == permission)
        query = self._build_query(
            query, agent, target, schema,
            Agent=Agent, Target=Target, custom=custom,
        )
        return bool(query.first())

    def _add_permission(self, agent, target, schema, permission):
        try:
            return schema.create(
                agent=agent,
                target=target,
                permission=permission,
            )
        except pw.IntegrityError:
            raise exceptions.PermissionExists()

    def _remove_permission(self, agent, target, schema, permission):
        query = schema.delete()
        query = query.where(schema.permission == permission)
        query = self._build_query(query, agent, target, schema)
        count = query.execute()
        if not count:
            raise exceptions.PermissionNotFound


def _reference_column(schema, **kwargs):
    return pw.ForeignKeyField(
        schema,
        on_update='CASCADE',
        on_delete='CASCADE',
        **kwargs
    )


class PeeweePermissionSchemaFactory(models.BasePermissionSchemaFactory):
    """Permission schema factory for use with Peewee.

    Note: Peewee does not implement delete and update cascades outside the
    database backend, so care should be taken with deleting agent and target
    records when using a backend that does not support cascades (e.g. SQLite,
    MySQL using the MyISAM storage engine). In this case, use the `recursive`
    flag when deleting through Peewee:

    .. code-block:: python

        agent.delete_instance(recursive=True)

    """
    @staticmethod
    def _get_table_name(schema):
        return schema._meta.db_table

    def _make_schema_meta(self, agent, target):
        return type(
            'Meta',
            (object, ),
            dict(
                db_table=self._make_table_name(agent, target),
                indexes=(
                    (('agent', 'target', 'permission'), True),
                ),
            ),
        )

    def _make_schema_dict(self, agent, target):
        return dict(
            Meta=self._make_schema_meta(agent, target),
            id=pw.PrimaryKeyField(),
            agent=_reference_column(agent, null=False, index=True),
            target=_reference_column(target, null=False, index=True),
            permission=pw.CharField(null=False, index=True),
        )


class PeeweeLoader(models.BaseLoader):

    def __init__(self, schema, column=None, kwarg='id'):
        column = column if column is not None else schema._meta.primary_key
        super(PeeweeLoader, self).__init__(schema, column, kwarg)

    def __call__(self, *args, **kwargs):
        return self.schema.select().where(
            self.column == kwargs.get(self.kwarg)
        ).first()



# -*- coding: utf-8 -*-

from guardrail.core import exceptions


class _Registry(object):
    """Registry of schemas designated as permission agents and targets, as well
    as the permission tables linking each agent-target pair. Should be used as
    a singleton in ordinary use, but can be instantiated for use in tests.
    """
    def __init__(self):
        self._agents = set()
        self._targets = set()
        self._permissions = dict()

    @property
    def agents(self):
        return self._agents

    @property
    def targets(self):
        return self._targets

    @property
    def permissions(self):
        return self._permissions.values()

    def agent(self, agent):
        """Decorator that registers the decorated schema as a permission agent.

        Example:

        .. code-block:: python

            @registry.agent
            class User(Base):
                id = sa.Column(sa.Integer, primary_key=True)

        """
        self._agents.add(agent)
        return agent

    def target(self, target):
        """Decorator that registers the decorated schema as a permission target.

        Example:

        .. code-block:: python

            @registry.target
            class Post(Base):
                id = sa.Column(sa.Integer, primary_key=True)

        """
        self._targets.add(target)
        return target

    def add_permission(self, agent, target, permission):
        """Get the join table linking schemas `agent` and `target`.

        :param agent: Agent schema class
        :param target: Target schema class
        :param schema: Permission join table
        """
        self._permissions[(agent, target)] = permission

    def get_permission(self, agent, target):
        """Get the join table linking schemas `agent` and `target`.

        :param agent: Agent schema class
        :param target: Target schema class
        :returns: Permission join table
        :raises: guardian.core.exceptions.SchemaNotFound if join table does not
            exist
        """
        try:
            return self._permissions[(agent, target)]
        except KeyError:
            raise exceptions.SchemaNotFound(
                'Could not find permission schema linking models {0} and {1}'.format(
                    agent.__name__,
                    target.__name__,
                )
            )

    def make_schemas(self, factory):
        """Create and register join tables linking all registered agent-target
        pairs.

        :param factory: Callable that takes agent and target schemas and returns
            the schema for a permission join table
        """
        for agent in self.agents:
            for target in self.targets:
                permission = factory(agent, target)
                self.add_permission(agent, target, permission)


registry = _Registry()



from django.conf.urls import patterns, include, url
from django.contrib import admin

urlpatterns = patterns('',
    # Examples:
    url(r'^$', 'djanguard.views.index', name='index'),
    url(r'^admin/', include(admin.site.urls)),
)



# -*- coding: utf-8 -*-

from flask.ext.sqlalchemy import SQLAlchemy

from guardrail.core import registry
from guardrail.ext.sqlalchemy import SqlalchemyPermissionSchemaFactory

db = SQLAlchemy()

@registry.agent
class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String, unique=True)

@registry.target
class Post(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    title = db.Column(db.String)
    content = db.Column(db.Text)

factory = SqlalchemyPermissionSchemaFactory((db.Model, ))
registry.make_schemas(factory)



# -*- coding: utf-8 -*-

class GuardrailException(Exception):
    pass


class RecordNotSaved(GuardrailException):
    pass


class SchemaNotFound(GuardrailException):
    pass


class PermissionExists(GuardrailException):
    pass


class PermissionNotFound(GuardrailException):
    pass



# -*- coding: utf-8 -*-
"""Abstract base classes for `guardrail` plugins. Plugin developers should
subclass :class:`BasePermissionManager` and :class:`BasePermissionSchemaFactory`,
and optionally :class:`BaseLoader` as well.
"""

import abc

import six

from guardrail.core import exceptions
from guardrail.core.registry import registry


def _get_class(value):
    return value if isinstance(value, type) else type(value)


@six.add_metaclass(abc.ABCMeta)
class BasePermissionManager(object):
    """Abstract base class for permission managers. Concrete subclasses must
    implement CRUD operations (check, add, and remove permissions), as well as
    the :meth:`_is_saved` check.

    :param _Registry registry: Optional registry object; use global `registry`
        if not provided.
    """
    def __init__(self, registry=registry):
        self.registry = registry

    def get_permissions(self, agent, target,
                        Agent=None, Target=None, custom=None):
        """List all permissions record `agent` has on record `target`.

        :param agent: Agent record
        :param target: Target record
        :param Agent: Optional agent schema; provide if permission is not
            directly related to `agent`
        :param Target: Optional target schema; provide if permission is not
            directly related to `target`
        :param custom: Optional callable for customizing permission queries; if
            provided, will be passed the original query, `agent`, `target`, and
            the permission join table
        :returns: Set of permission labels between `agent` and `target`
        """
        schema = self._get_permission_schema(agent, target, Agent, Target)
        return self._get_permissions(
            agent, target, schema,
            Agent=Agent, Target=Target, custom=custom
        )

    def has_permission(self, agent, target, permission,
                       Agent=None, Target=None, custom=None):
        """Check whether record `agent` has permission `permission` on record
        `target`.

        :param agent: Agent record
        :param target: Target record
        :param str permission: Permission
        :param Agent: Optional agent schema; provide if permission is not
            directly related to `agent`
        :param Target: Optional target schema; provide if permission is not
            directly related to `target`
        :param custom: Optional callable for customizing permission queries; if
            provided, will be passed the original query, `agent`, `target`, and
            the permission join table
        :returns: Record `agent` has permission `permission` on record `target`
        """
        schema = self._get_permission_schema(agent, target, Agent, Target)
        return self._has_permission(
            agent, target, schema, permission,
            Agent=Agent, Target=Target, custom=custom,
        )

    def add_permission(self, agent, target, permission):
        """Grant permission `permission` to record `agent` on record `target`.

        :param agent: Agent record
        :param target: Target record
        :param str permission: Permission
        :returns: Created permission record
        :raises: `PermissionExistsError` if permission has already been granted
        """
        schema = self._get_permission_schema(agent, target)
        return self._add_permission(agent, target, schema, permission)

    def remove_permission(self, agent, target, permission):
        """Revoke permission `permission` from record `agent` on record `target`.

        :param agent: Agent record
        :param target: Target record
        :param str permission: Permission
        :raises: `PermissionNotFound` if permission has not been granted
        """
        schema = self._get_permission_schema(agent, target)
        return self._remove_permission(agent, target, schema, permission)

    def _get_permission_schema(self, agent, target, Agent=None, Target=None):
        """Look up join table linking `agent` and `target`, verifying that both
        records have been persisted.

        :param agent: Agent record
        :param target: Target record
        :param Agent: Optional agent schema; provide if permission is not
            directly related to `agent`
        :param Target: Optional target schema; provide if permission is not
            directly related to `target`
        :returns: Join table between `agent` and `target`
        :raises: `RecordNotSaved` if either record has not been persisted
        :raises: `SchemaNotFound` if no join table exists
        """
        self._check_saved(agent, target)
        return self.registry.get_permission(
            _get_class(Agent or agent),
            _get_class(Target or target),
        )

    def _check_saved(self, *records):
        for record in records:
            if not self._is_saved(record):
                raise exceptions.RecordNotSaved('Record {0!r} not saved'.format(record))

    @abc.abstractmethod
    def _is_saved(self, record):
        """Check whether `record` has been persisted. Note: For backends that
        do not require records to be persisted to operate on foreign-key
        relationships, this method should always return `True`.

        :param record: Record to check
        :returns: `record` has been persisted
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def _get_permissions(self, agent, target, schema):
        pass  # pragma: no cover

    @abc.abstractmethod
    def _has_permission(self, agent, target, schema, permission):
        pass  # pragma: no cover

    @abc.abstractmethod
    def _add_permission(self, agent, target, schema, permission):
        pass  # pragma: no cover

    @abc.abstractmethod
    def _remove_permission(self, agent, target, schema, permission):
        pass  # pragma: no cover


@six.add_metaclass(abc.ABCMeta)
class BasePermissionSchemaFactory(object):
    """Abstract base class for schema factories that create permission join
    tables. Concrete subclasses must implement :meth:`_get_table_name` and
    :meth:`_make_schema_dict`.

    :param tuple bases: Base classes for created schema classes
    """
    def __init__(self, bases):
        self.bases = bases

    def __call__(self, agent, target):
        """Create a join table representing permissions between `agent` and
        `target` schemas.

        :param agent: Agent schema class
        :param target: Target schema class
        :returns: Created schema class
        """
        schema = type(
            self._make_schema_name(agent, target),
            self.bases,
            self._make_schema_dict(agent, target),
        )
        self._update_parents(agent, target, schema)
        return schema

    def _update_parents(self, agent, target, schema):
        """Creating a permission join table may require mutating the `agent`
        and `target` schemas. By default, take no action.

        :param agent: Agent schema class
        :param target: Target schema class
        :param schema: Created schema class
        """
        pass

    def _make_schema_name(self, agent, target):
        """Build class name for permission join table.

        :param agent: Agent schema class
        :param target: Target schema class
        """
        return '{0}{1}Permission'.format(
            agent.__name__,
            target.__name__,
        )

    def _make_table_name(self, agent, target):
        """Build table name for permission join table.

        :param agent: Agent schema class
        :param target: Target schema class
        """
        return '{0}_{1}_permission'.format(
            self._get_table_name(agent),
            self._get_table_name(target),
        )

    @abc.abstractmethod
    def _get_table_name(self, schema):
        """Get table name for `schema`.

        :param schema: Schema class
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def _make_schema_dict(self, agent, target):
        """Build class dictionary for permission join table.

        :param agent: Agent schema class
        :param target: Target schema class
        :returns: Dictionary of class members
        """
        pass  # pragma: no cover


class BaseLoader(object):

    def __init__(self, schema, column, kwarg='id'):
        self.schema = schema
        self.column = column
        self.kwarg = kwarg

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        pass  # pragma: no cover



# -*- coding: utf-8 -*-

import re
import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand


REQUIREMENTS = [
    'six',
]
TEST_REQUIREMENTS = [
    'pytest',
    'pytest-cov',
    'pytest-django',
    'pytest-pythonpath',
    'peewee',
    'pony',
    'Django',
    'SQLAlchemy',
]


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = [
            '--cov', 'guardrail', 'tests',
            '--cov-report', 'term-missing',
        ]
        self.test_suite = True

    def run_tests(self):
        import pytest
        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


def find_version(fname):
    """Attempts to find the version number in the file names fname.
    Raises RuntimeError if not found.
    """
    version = ''
    with open(fname, 'r') as fp:
        reg = re.compile(r'__version__ = [\'"]([^\'"]*)[\'"]')
        for line in fp:
            m = reg.match(line)
            if m:
                version = m.group(1)
                break
    if not version:
        raise RuntimeError('Cannot find version information')
    return version


__version__ = find_version('guardrail/__init__.py')


def read(fname):
    with open(fname) as fp:
        content = fp.read()
    return content


setup(
    name='guardrail',
    version=__version__,
    description=(
        'An extensible library for managing object-level permissions with '
        'support for SQLAlchemy, Peewee, Pony, and Django.'
    ),
    long_description=read('README.rst'),
    author='Joshua Carp',
    author_email='jm.carp@gmail.com',
    url='https://github.com/jmcarp/guardrail',
    packages=find_packages(exclude=('test*', 'examples')),
    package_dir={'guardrail': 'guardrail'},
    install_requires=REQUIREMENTS,
    license=read('LICENSE'),
    ip_safe=False,
    keywords=(
        'guardrail', 'authorization', 'sql',
        'sqlalchemy', 'peewee', 'pony', 'django',
    ),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    test_suite='tests',
    tests_require=TEST_REQUIREMENTS,
    cmdclass={'test': PyTest},
)
